#!/usr/bin/perl -w
##
# Execution system for the DCPU-16 processor.
#
# This is a pretty simple interpreter for the core data.
#
# We work on the principle that all operands are references. If an operand
# is read, we use the 'readref', and if the operand is written to we use
# the 'writeref'. The references may be to registers, or to core memory
# locations. They may also be objects which implement hardware, in which
# case their 'read' or 'write' functions are called.
#
# Consequently some of the opcodes are implemented oddly. Where possible
# we try to keep to the same semantics such that any memory accesses will
# obey hardware mappings that might be present.
#
# Actual hardware regions will be implemented by separate objects, and
# will be registered as part of hardware profiles.
#
# Typical use of this class might be:
# 
#    $dasm->readfile($infile);  # Assemble some code (or load a dump)
# 
#    # Resolve all the relocations
#    die "Unresolved relocations" if (!$dasm->resolve($infile));
#
#    $dexec = new DExec($dasm);
#    while (1)
#    {
#        $dexec->runone();
#    }
#
#
# @file
# @author Justin Fletcher
#

package DExec;

use DCPU;

use DExecHWAdapter;
use DExecHW::DebugNumber;

## The indices of the registers in our 'registers' array.
our $O  = 8;
our $SP = 9;
our $PC = 10;
our @RegisterNames = (
        (sort { $DCPU::RegisterMap{$a} <=> $DCPU::RegisterMap{$b} }
            keys %DCPU::RegisterMap),  # 0-7
        'O',                        # 8
        'SP',                       # 9
        'PC',                       # 10
    );


# Whether we want to debug the reference handling
my $debug_ref = 0;

##
# The functions to apply for every basic opcode.
# 
# Each function is passed the DExec object, operand a and operand b.
my %basicFuncs = (

    # 0x1: SET a, b - sets a to b
    $DCPU::BasicOps{'SET'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            $self->writeref($operand_a, $self->readref($operand_b));
            $self->cycles(1);
        },

    # 0x2: ADD a, b - sets a to a+b, sets O to 0x0001 if there's
    #      an overflow, 0x0 otherwise
    $DCPU::BasicOps{'ADD'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a + $b;
            
            $self->writeref($operand_a, $value);
            $self->{'registers'}->[ $O ] = ($value >> 16) & 0xFFFF;
            $self->cycles(2);
        },
    
    # 0x3: SUB a, b - sets a to a-b, sets O to 0xffff if there's an
    #      underflow, 0x0 otherwise
    $DCPU::BasicOps{'SUB'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a - $b;
            
            $self->writeref($operand_a, $value);
            $self->{'registers'}->[ $O ] = ($value >> 16) & 0xFFFF;
            $self->cycles(2);
        },

    # 0x4: MUL a, b - sets a to a*b, sets O to ((a*b)>>16)&0xffff
    $DCPU::BasicOps{'MUL'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a * $b;
            
            $self->writeref($operand_a, $value);
            $self->{'registers'}->[ $O ] = ($value >> 16) & 0xFFFF;
            $self->cycles(2);
        },

    # 0x5: DIV a, b - sets a to a/b, sets O to ((a<<16)/b)&0xffff.
    #      if b==0, sets a and O to 0 instead.
    $DCPU::BasicOps{'DIV'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            if ($b == 0)
            {
                $self->writeref($operand_a, 0);
                $self->{'registers'}->[ $O ] = 0;
            }
            else
            {
                my $value = $a / $b;

                $self->writeref($operand_a, int($value));
                $self->{'registers'}->[ $O ] = ($value << 16) & 0xFFFF;
            }
            $self->cycles(3);
        },
        
    # 0x6: MOD a, b - sets a to a%b. if b==0, sets a to 0 instead.
    $DCPU::BasicOps{'MOD'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            if ($b == 0)
            {
                $self->writeref($operand_a, 0);
            }
            else
            {
                my $value = $a % $b;

                $self->writeref($operand_a, int($value));
            }
            $self->cycles(3);
        },

    # 0x7: SHL a, b - sets a to a<<b, sets O to ((a<<b)>>16)&0xffff
    $DCPU::BasicOps{'SHL'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a << $b;
            
            $self->writeref($operand_a, $value);
            $self->{'registers'}->[ $O ] = ($value >> 16) & 0xFFFF;
            $self->cycles(2);
        },

    # 0x8: SHR a, b - sets a to a>>b, sets O to ((a<<16)>>b)&0xffff
    $DCPU::BasicOps{'SHR'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a >> $b;
            
            $self->writeref($operand_a, $value);
            $self->{'registers'}->[ $O ] = (($a<<16) >> $b) & 0xFFFF;
            $self->cycles(2);
        },

    # 0x9: AND a, b - sets a to a&b
    $DCPU::BasicOps{'AND'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a & $b;
            
            $self->writeref($operand_a, $value);
            $self->cycles(1);
        },

    # 0xa: BOR a, b - sets a to a|b
    $DCPU::BasicOps{'BOR'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a | $b;
            
            $self->writeref($operand_a, $value);
            $self->cycles(1);
        },

    # 0xb: XOR a, b - sets a to a^b
    $DCPU::BasicOps{'XOR'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            my $value = $a ^ $b;
            
            $self->writeref($operand_a, $value);
            $self->cycles(1);
        },

    # 0xc: IFE a, b - performs next instruction only if a==b
    $DCPU::BasicOps{'IFE'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            $self->{'executenext'} = ($a == $b);
            $self->cycles(2 + ($self->{'executenext'} ? 0 : 1));
        },

    # 0xd: IFN a, b - performs next instruction only if a!=b
    $DCPU::BasicOps{'IFN'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            $self->{'executenext'} = ($a != $b);
            $self->cycles(2 + ($self->{'executenext'} ? 0 : 1));
        },

    # 0xe: IFG a, b - performs next instruction only if a>b
    $DCPU::BasicOps{'IFG'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            $self->{'executenext'} = ($a > $b);
            $self->cycles(2 + ($self->{'executenext'} ? 0 : 1));
        },

    # 0xf: IFB a, b - performs next instruction only if (a&b)!=0
    $DCPU::BasicOps{'IFB'} => sub {
            my ($self, $operand_a, $operand_b) = @_;

            my $a = $self->readref($operand_a);
            my $b = $self->readref($operand_b);
            $self->{'executenext'} = ($a & $b) != 0;
            $self->cycles(2 + ($self->{'executenext'} ? 0 : 1));
        },
);


##
# The functions to apply for every extended opcode.
# 
# Each function is passed the DExec object, and operand a.
my %extendedFuncs = (

    # 0x01: JSR a - pushes the address of the next instruction to the
    # stack, then sets PC to a
    $DCPU::ExtendedOps{'JSR'} => sub {
            my ($self, $operand_a) = @_;
            
            $self->writeref( $self->PUSH(), $self->{'registers'}->[ $PC ]);

            $self->{'registers'}->[ $PC ] = $self->readref($operand_a);
            $self->cycles(2);
        },
);


##
# Construct a new assembler object.
#
# @param[in]  $proto   This class object, or prototype to add to
# @param[in]  $core    The core to execute (may be a hashref
#                      of address=>value, or DASM object reference)
# @param[in]  $pc      Address to execute at, defaults to 0
#
# @return Assembler object
sub new
{
    my $proto = shift;
    my $core = shift;
    my $pc = shift || 0;
    my $class = ref($proto) || $proto;
    my $self;
    
    if (!defined $core)
    {
        $core = {};
    }
    elsif (ref $core eq 'DASM')
    {
        $core = $core->{'core'};
    }

    $self = {
            # Core values
            'core' => $core,
            
            # Register values
            'registers' => [ 
                    map {
                        0
                    } (@RegisterNames)
                ],
            
            # 'Execute next instruction' flag
            'executenext' => 1,

            # Cycle counter, incremented by each instruction
            'cycles' => 0,

            # Instructions executed
            'instructions' => 0,
        };
    
    $self->{'registers'}->[ $PC ] = $pc;
    
    bless $self, $class;
    
    return $self;
}


##
# Increment the number of cycles taken
#
# @param[in] $self   DExec object
# @param[in] $cycles Number of cycles consumed, or undef to read
#
# @return Total number of cycles.
sub cycles
{
    my ($self, $cycles) = @_;
    
    $self->{'cycles'} += $cycles if (defined $cycles);
    
    return $self->{'cycles'};
}


##
# Return the number of instructions executed
#
# @param[in] $self   DExec object
#
# @return Total number of instructions
sub instructions
{
    my ($self) = @_;
    
    return $self->{'instructions'};
}


## 
# Get a word from the core.
#
# Any core values that haven't been assigned will return 0.
#
# @param[in] $self   DExec object
# @param[in] $addr   Address to read from
#
# @return value from the core
sub getword
{
    my ($self, $addr) = @_;
    
    $addr = $addr & 0xffff;
    
    if (defined $self->{'core'}->{ $addr })
    {
        return $self->{'core'}->{ $addr };
    }
    else
    {
        return 0;
    }
}


## 
# Get a reference to a word from the core.
#
# Any core values that haven't been assigned will be assigned 0.
#
# @param[in] $self   DExec object
# @param[in] $addr   Address to read from
#
# @return reference to a value from the core
sub getwordref
{
    my ($self, $addr) = @_;
    
    $addr = $addr & 0xffff;
    
    if (!defined $self->{'core'}->{ $addr })
    {
        $self->{'core'}->{ $addr } = 0;

    }
    my $ref = \$self->{'core'}->{ $addr };
    
    printf "> getwordref: %s = %04x\n", $ref, $addr if ($debug_ref);
    
    return $ref;
}


##
# Read PC and increment.
#
# @param[in] $self   DExec object
#
# @return value from the core and increment PC
sub nextword
{
    my ($self) = @_;
    
    my $pc = $self->{'registers'}->[ $PC ]++;
    
    return $self->getword($pc);
}


##
# Get the value of an operand, given the operand value.
#
# @param[in] $self      DExec object
# @param[in] $operand   Operand value
#
# @return $valref   reference to the operand value, or an
#                   object to access value
# @note Constants will still be returned as a reference (to a value
#       that goes nowhere)
sub getoperand
{
    my ($self, $operand) = @_;
    
    # 0x00-0x07: register (A, B, C, X, Y, Z, I or J, in that order)
    if ($operand >= 0x00 && $operand <= 0x07)
    {
        return \$self->{'registers'}->[ $operand ];
    }
    # 0x08-0x0f: [register]
    elsif ($operand >= 0x08 && $operand <= 0x0f)
    {
        $self->cycles(1) if ($self->{'executenext'}); # 1 extra cycle for the read
        return $self->getwordref( $self->{'registers'}->[ $operand ]);
    }
    # 0x10-0x17: [next word + register]
    elsif ($operand >= 0x10 && $operand <= 0x17)
    {
        my $nextword = $self->nextword();
        # an extra cycles for the read of the referenced value
        $self->cycles(1) if ($self->{'executenext'});
        return $self->getwordref( $nextword +
                                  $self->{'registers'}->[ $operand - 0x10] );
    }
    #      0x18: POP / [SP++]
    elsif ($operand == 0x18)
    {
        $self->cycles(1) if ($self->{'executenext'}); # 1 extra cycle for the read
        return $self->POP();
    }
    #      0x19: PEEK / [SP]
    elsif ($operand == 0x19)
    {
        $self->cycles(1) if ($self->{'executenext'}); # 1 extra cycle for the read
        return $self->PEEK();
    }
    #      0x1a: PUSH / [--SP]
    elsif ($operand == 0x1a)
    {
        $self->cycles(1) if ($self->{'executenext'}); # 1 extra cycle for the read
        return $self->PUSH();
    }
    #      0x1b: SP
    elsif ($operand == 0x1b)
    {
        return \$self->{'registers'}->[ $SP ];
    }
    #      0x1c: PC
    elsif ($operand == 0x1c)
    {
        return \$self->{'registers'}->[ $PC ];
    }
    #      0x1d: O
    elsif ($operand == 0x1d)
    {
        return \$self->{'registers'}->[ $O ];
    }
    #      0x1e: [next word]
    elsif ($operand == 0x1e)
    {
        # 1 extra cycles for the referenced value
        $self->cycles(1) if ($self->{'executenext'});
        return $self->getwordref( $self->nextword() );
    }
    #      0x1f: next word (literal)
    elsif ($operand == 0x1f)
    {
        my $pc = $self->{'registers'}->[ $PC ]++;
        return $self->getwordref($pc);
    }
    # 0x20-0x3f: literal value 0x00-0x1f (literal)
    elsif ($operand >= 0x20 && $operand <= 0x3f)
    {
        my $value = $operand - 0x20;
        return \$value;
    }
    else
    {
        die "Invalid operand value $operand";
    }
}


##
# Read a value from a reference.
#
# @param[in] $self   DExec object
# @param[in] $ref    Reference to read the value of
#
# @return value
sub readref
{
    my ($self, $ref) = @_;
    
    if (ref $$ref ne '')
    {
        # This is a hardware implementation, so invoke the writer
        print "> ReadRefHW: ".$$ref->describe()."\n" if ($debug_ref);
        return $$ref->read($exec, $value);
    }
    else
    {
        print "> ReadRef: $ref\n" if ($debug_ref);
        return $$ref;
    }
}


##
# Write a value to a reference.
#
# @param[in] $self   DExec object
# @param[in] $ref    Reference to write to
# @param[in] $value  Value to write
sub writeref
{
    my ($self, $ref, $value) = @_;
    
    die "Attempted to write an undef\n" if (!defined $value); 
    
    $value = $value & 0xFFFF;
    
    if (ref $$ref ne '')
    {
        # This is a hardware implementation, so invoke the writer
        print "> WriteRefHW: ".$$ref->describe()." : $value\n" if ($debug_ref);
        $$ref->write($exec, $value);
    }
    else
    {
        print "> WriteRef: $ref : $value\n" if ($debug_ref);
        $$ref = $value;
    }
}



##
# POP (unstack a value) from the full descending stack
#
# @param[in] $self   DExec object
#
# @return reference to the value popped
sub POP
{
    my ($self) = @_;
    my $sp = $self->{'registers'}->[ $SP ]++;
    if ($self->{'registers'}->[ $SP ] > 0xFFFF)
    {
        $self->{'registers'}->[ $SP ] = 0;
    }
    return $self->getwordref($sp);
}


##
# PUSH (stack a value) on to the full descending stack
#
# @param[in] $self   DExec object
#
# @return reference to the value popped
sub PUSH
{
    my ($self) = @_;
    $self->{'registers'}->[ $SP ] -= 1;
    if ($self->{'registers'}->[ $SP ] < 0)
    {
        $self->{'registers'}->[ $SP ] = 0xFFFF;
    }
    return $self->getwordref( $self->{'registers'}->[ $SP ] );
}


##
# PEEK (look at the stack value) in a full descending stack
#
# @param[in] $self   DExec object
#
# @return reference to the value popped
sub PEEK
{
    my ($self) = @_;
    my $sp = $self->{'registers'}->[ $SP ];
    return $self->getwordref($sp);
}



##
# Run a single instruction.
#
# @param[in] $self   DExec object
sub runone
{
    my ($self) = @_;
    
    my $instruction = $self->nextword();
    
    my $opcode = ($instruction & 15);
    
    if ($opcode != 0)
    {
        # A basic opcode
        my $op = $basicFuncs{$opcode};
        my $pc = $self->{'registers'}->[ $PC ];
        my $operand_a = $self->getoperand(($instruction >> 4) & 0x3f);
        my $operand_b = $self->getoperand(($instruction >> 10) & 0x3f);
        my $extracycles = $self->{'registers'}->[ $PC ] - $pc;
        
        if ($self->{'executenext'})
        {
            if (!defined $op)
            {
                # This should never happen
                die "Attempt to execute invalid opcode $opcode";
            }
            
            # Increment our counter
            $self->{'instructions'}++;
            
            $self->{'cycles'} += $extracycles;

            # Run the instruction
            & $op ($self, $operand_a, $operand_b);
        }
        else
        {
            $self->{'executenext'} = 1;
        }
    }
    else
    {
        # Opcode 0 is an extended opcode
        $opcode = ($instruction >> 4) & 0x3f;
        my $op = $extendedFuncs{$opcode};
        
        my $operand_a = $self->getoperand(($instruction >> 10) & 0x3f);
        
        if ($self->{'executenext'})
        {
            if (!defined $op)
            {
                # This should never happen
                die "Attempt to execute invalid extended opcode $opcode";
            }
            
            # Increment our counter
            $self->{'instructions'}++;

            # Run the instruction
            & $op ($self, $operand_a);
        }
        else
        {
            $self->{'executenext'} = 1;
        }
    }
}


##
# Show the current registers state
#
# The registers will be listed, followed by value of the instruction
# at PC, and 'NX' if the 'no execute' flag has been set for the next
# instruction.
# Following the registers, the stack will be printed, ascending from
# SP to the top of memory.
#
# @param[in]  $self     DExec object
# @param[in]  $prefix   Prefix for each line
sub showregs
{
    my ($self, $prefix) = @_;
    $prefix = "" if (!defined $prefix);
    
    print "$prefix";
    for my $i (0..$#RegisterNames)
    {
        printf "%2s: %04x ", $RegisterNames[$i],
                              $self->{'registers'}->[$i];
        if (($i & 7) == 7)
        {
            print "\n$prefix";
        }
    }
    printf "[%04x]  ", $self->getword( $self->{'registers'}->[ $PC ] );

    if (!$self->{'executenext'})
    {
        print "NX";
    }
    
    print "\n";
 
    # Show the stack
    my $sp = $self->{'registers'}->[$SP];
    print "${prefix}STK> ";
    my $wrote = 0;
    while ($sp != 0 && $sp < 0x10000)
    {
        printf "[%04x]  ", $self->getword( $sp );
        $sp++;
        $wrote++;
        if (($wrote & 16)==15 && $sp != 0x10000)
        {
            print "\n$prefix     ";
        }
    }
    print "\n\n";
}


##
# Read a named register.
#
# @param[in] $self   Dexec object
# @param[in] $reg    named register
#
# @return register value
sub getreg
{
    my ($self, $reg) = @_;
    
    $reg = uc $reg;
    
    if (defined $DCPU::RegisterMap{$reg})
    {
        return $self->{'registers'}->[$DCPU::RegisterMap{$reg}];
    }
    elsif ($reg eq 'PC')
    {
        return $self->{'registers'}->[$PC];
    }
    elsif ($reg eq 'SP')
    {
        return $self->{'registers'}->[$SP];
    }
    elsif ($reg eq 'O')
    {
        return $self->{'registers'}->[$O];
    }
    else
    {
        die "Unrecognised register $reg\n";
    }
}

# Must return true
1;
