#!/usr/bin/perl -w
##
# HW implementation for 'SimpleTimer'.
#
# A very simple timer measuring microseconds since the hardware was
# initialised. FIXME Allow this to be CPU start?
#
# 3 words are provided (little endian):
#    +0 low word    (bits 0-15)
#    +1 middle word (bits 16-31)
#    +2 high word   (bits 32-47)
#
# The middle and high words are latched by reading the low word.
#
# Writing to any register has the effect of resetting the clock to
# 0.
#

package DExecHW::SimpleTimer;

use Time::HiRes qw(gettimeofday tv_interval);


##
# Construct a new hardware device object.
#
# @param[in]  $proto   This class object, or prototype to add to
# @param[in]  $object  The object to proxy for
# @param[in]  $offset  The offset this object is created for
#
# @return Assembler object
sub new
{
    my $proto = shift;
    my $object = shift;
    my $offset = shift || 0;
    my $class = ref($proto) || $proto;
    my $self = {
            'initial' => [gettimeofday],
            0         => 0,
            1         => 0,
            2         => 0,
        };

    bless $self, $class;
    
    return $self;
}


##
# Return the size of of the implementation.
#
# @param[in] $self   Object
#
# @return Size of the hardware implementation in words
sub window
{
    # 3 words is 2^48 microseconds, which is 3257 days.
    return 3;
}


##
# Read value
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
# @param[in] $exec   DExec object opon which this object is operating
#
# @return value of this register
sub read
{
    my ($self, $offset, $exec) = @_;
    
    if ($offset == 0)
    {
        # Latch value into registers
        my $diff = tv_interval($self->{'initial'});
        
        # Measure in milliseconds.
        $diff = $diff * 1E6;
        
        $self->{0} = $diff & 0xFFFF;
        $self->{1} = ($diff>>16) & 0xFFFF;
        $self->{2} = ($diff>>32) & 0xFFFF;
    }
    
    return $self->{$offset};
}


##
# Write value
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
# @param[in] $exec   DExec object upon which this object is operating
# @param[in] $value  Value to write
#
# @return ignored
sub write
{
    my ($self, $offset, $exec, $value) = @_;
    
    $self->{'initial'} = [gettimeofday];
}


##
# Get the symbol name for an offset
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
#
# @return symbol name to use
# @retval undef for a default symbol
# @retval "" for no symbol
sub symbol
{
    my ($self, $offset) = @_;
    
    return ("Low", "Middle", "High")[$offset];
}


# Must return 1
1;
